<?php
/**
 * YetiForce system update package file.
 *
 * @package   YetiForce.UpdatePackages
 *
 * @copyright YetiForce S.A.
 * @license   YetiForce Public License 7.0 (licenses/LicenseEN.txt or yetiforce.com)
 * @author    Radosław Skrzypczak <r.skrzypczak@yetiforce.com>
 */

/**
 * YetiForce system update package class.
 */
class YetiForceUpdate
{
    /** @var \vtlib\PackageImport */
    public $package;

    /** @var string[] Fields to delete. */
    public $filesToDelete = [];

    /** @var string */
    private $logFile = 'cache/logs/updateLogsTrace.log';

    /** @var object Module Meta XML File (Parsed). */
    private $moduleNode;

    /** @var DbImporter */
    private $importer;

    /** @var string[] Errors. */
    private $error = [];

    /**
     * Constructor.
     *
     * @param object $moduleNode
     */
    public function __construct($moduleNode)
    {
        $this->moduleNode = $moduleNode;
        $this->filesToDelete = require_once 'deleteFiles.php';
    }

    /**
     * Log.
     *
     * @param string $message Logs.
     */
    private function log(string $message): void
    {
        $fp = fopen($this->logFile, 'a+');
        fwrite($fp, $message . PHP_EOL);
        fclose($fp);
        if (false !== stripos($message, '[ERROR]')) {
            $this->error[] = $message;
        }
    }

    /**
     * Pre update.
     */
    public function preupdate(): bool
    {
        $start = microtime(true);
        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s'));

        $minTime = 600;
        $maxExecutionTime = ini_get('max_execution_time');
        $maxInputTime = ini_get('max_input_time');

        if (version_compare(PHP_VERSION, '8.1', '<')) {
            $this->package->_errorText = 'The server configuration is not compatible with the requirements of the upgrade package.' . PHP_EOL;
            $this->package->_errorText .= 'Please have a look at the list of errors:' . PHP_EOL . PHP_EOL;
            $this->package->_errorText .= 'Wrong PHP version, recommended version >= 8.1';
            $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s') . ' | ' . round((microtime(true) - $start) / 60, 2) . ' min');
            return false;
        }

        if ((is_numeric($maxExecutionTime) && 0 != $maxExecutionTime && $maxExecutionTime < $minTime) || (is_numeric($maxInputTime) && $maxInputTime > 0 && $maxInputTime < $minTime)) {
            $this->package->_errorText = 'The server configuration is not compatible with the requirements of the upgrade package.' . PHP_EOL;
            $this->package->_errorText .= 'Please have a look at the list of errors:';
            if (0 != $maxExecutionTime && $maxExecutionTime < $minTime) {
                $this->package->_errorText .= PHP_EOL . 'max_execution_time = ' . $maxExecutionTime . ' < ' . $minTime;
            }
            if ($maxInputTime > 0 && $maxInputTime < $minTime) {
                $this->package->_errorText .= PHP_EOL . 'max_input_time = ' . $maxInputTime . ' < ' . $minTime;
            }
            $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s') . ' | ' . round((microtime(true) - $start) / 60, 2) . ' min');
            return false;
        }

        if(!\function_exists('openssl_encrypt') || !extension_loaded('openssl')) {
            $this->package->_errorText = 'The server configuration is not compatible with the requirements: openssl extension is required.' . PHP_EOL;
            return false;
        }
        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s') . ' | ' . round((microtime(true) - $start) / 60, 2) . ' min');

        return true;
    }

    /**
     * Update.
     */
    public function update(): void
    {
        $start = microtime(true);
        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s'));

        $this->importer = new \App\Db\Importer();
        try {
            $this->importer->loadFiles(__DIR__ . '/dbscheme');
            $this->importer->checkIntegrity(false);

            $this->importer->updateScheme();
            $this->importer->refreshSchema();
            $this->importer->logs(false);
        } catch (\Throwable $ex) {
            $this->log($ex->getMessage() . '|' . $ex->getTraceAsString());
            $this->importer->logs(false);
            $this->importer->checkIntegrity(true);
            throw $ex;
        }
        $this->importer->checkIntegrity(true);

        $this->updateData();
        $this->createConfigFiles();


        $this->importer->logs(false);
        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s') . ' | ' . round((microtime(true) - $start) / 60, 2) . ' min');
    }

    private function updateData(): void
    {
        $start = microtime(true);
        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s'));

        $updates = [
            ['vtiger_field', ['presence' => 2], ['columnname' => 'account_short_name', 'tablename' => 'vtiger_account', 'presence' => 0]],
        ];
        $batchUpdate = \App\Db\Updater::batchUpdate($updates);
        $this->log('  [INFO] batchUpdate: ' . \App\Utils::varExport($batchUpdate));
        unset($batchUpdate);

        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s') . ' | ' . round((microtime(true) - $start) / 60, 2) . ' mim.');
    }

    /**
     * Post update.
     */
    public function createConfigFiles(): bool
    {
        $start = microtime(true);
        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s'));

        $files = ['config/ConfigTemplates.php'];
        foreach ($files as $configTemplates) {
            if(!file_exists(__DIR__ . '/files/' . $configTemplates)) {
                continue;
            }
            $path = ROOT_DIRECTORY . '/' . $configTemplates;
            copy(__DIR__ . '/files/' . $configTemplates, $path);
            \App\Cache::resetFileCache($path);
        }

        \App\Cache::resetOpcache();
        clearstatcache();

        $configFile = new \App\ConfigFile('debug');
        $configFile->set('SMARTY_ERROR_REPORTING', new \Nette\PhpGenerator\Literal('E_ALL & ~E_NOTICE & ~E_DEPRECATED & ~E_USER_DEPRECATED'));
        $configFile->create();

        (new \App\ConfigFile('api'))->create();

        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s') . ' | ' . round((microtime(true) - $start) / 60, 2) . ' mim.');
        return true;
    }

    /**
     * Post update .
     */
    public function postupdate()
    {
        $start = microtime(true);
        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s'));

        (new \App\BatchMethod(['method' => '\App\Db\Fixer::baseModuleTools', 'params' => []]))->save();
        (new \App\BatchMethod(['method' => '\App\Db\Fixer::baseModuleActions', 'params' => []]))->save();
        (new \App\BatchMethod(['method' => '\App\Db\Fixer::profileField', 'params' => []]))->save();
        (new \App\BatchMethod(['method' => '\App\UserPrivilegesFile::recalculateAll', 'params' => []]))->save();
        (new \App\BatchMethod(['method' => 'Settings_SharingAccess_Module_Model::recalculateSharingRules', 'params' => []]))->save();
        (new \App\BatchMethod(['method' => '\App\Cache:::clearAll', 'params' => []]))->save();

        \App\Cache::clear();
        \App\Cache::resetOpcache();
        \App\Cache::clearAll();

        $db= \App\Db::getInstance();
        $db->createCommand()->insert('yetiforce_updates', [
            'user' => \App\User::getCurrentUserModel()->getDetail('user_name'),
            'name' => (string) $this->moduleNode->label,
            'from_version' => (string) $this->moduleNode->from_version,
            'to_version' => (string) $this->moduleNode->to_version,
            'result' => true,
            'time' => date('Y-m-d H:i:s'),
        ])->execute();

        $db->createCommand()->update('vtiger_version', ['current_version' => (string) $this->moduleNode->to_version])->execute();

        \vtlib\Functions::recurseDelete('cache/updates/updates');
        register_shutdown_function(function () {
            try {
                \vtlib\Functions::recurseDelete('cache/templates_c');
            } catch (\Exception $e) {
                \App\Log::error($e->getMessage() . PHP_EOL . $e->__toString());
            }
        });

        \vtlib\Functions::recurseDelete('app_data/LanguagesUpdater.json');
        \vtlib\Functions::recurseDelete('app_data/SystemUpdater.json');
        \vtlib\Functions::recurseDelete('app_data/cron.php');
        \vtlib\Functions::recurseDelete('app_data/ConfReport_AllErrors.php');

        file_put_contents('cache/logs/update.log', PHP_EOL . date('Y-m-d H:i:s') . ' | ' . ob_get_clean(), FILE_APPEND);
        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s') . ' | ' . round((microtime(true) - $start) / 60, 2) . ' min');
        if(PHP_SAPI !== 'cli') {
            header('location: index.php?parent=Settings&module=Updates&view=Index');
        }

        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s') . ' | ' . round((microtime(true) - $start) / 60, 2) . ' min');

        return exit;
    }

}
